"""
RT130 object to hold and process RefTek data
"""
import os
from pathlib import Path
from typing import Tuple, List, Union
import traceback
import numpy as np

from sohstationviewer.model.reftek.from_rt2ms.core import Reftek130
from sohstationviewer.model.reftek.from_rt2ms import (
    core, soh_packet, packet)
from sohstationviewer.model.reftek.log_info import LogInfo
from sohstationviewer.model.data_type_model import (
    DataTypeModel, ThreadStopped, ProcessingDataError)
from sohstationviewer.model.handling_data import read_text
from sohstationviewer.model.handling_data_reftek import (
    check_reftek_header, read_reftek_stream)
from sohstationviewer.conf import constants
from sohstationviewer.controller.util import validate_file
from sohstationviewer.view.util.enums import LogType


class RT130(DataTypeModel):
    """
    read and process reftek file into object with properties can be used to
    plot SOH data, mass position data, waveform data and gaps
    """
    def __init__(self, *args, **kwarg):
        self.EH = {}
        super().__init__(*args, **kwarg)
        """
        req_data_streams: list of data streams requested by user is kept
            in self.req_wf_chans to be compliant with mseed when passing to
            object. But assign to this name to be closer to actual meaning.
        """
        self.req_data_streams: List[Union[int, str]] = self.req_wf_chans
        """
        rt130_waveform_data_req: flag to create waveform data according to
            req_data_stream
        """
        self.rt130_waveform_data_req: bool = kwarg['rt130_waveform_data_req']
        """
        found_data_streams: list of data streams found to help inform user
            why the selected data streams don't show up
        """
        self.found_data_streams: List[int] = []
        self.processing_data()

    def finalize_data(self):
        """
        This function should be called after all folders finish reading to
            + create soh_data from log_data
            + check not found data stream to give user a warning if needed
            + other tasks in super().finalize_data()
        """
        self.track_info(
            "Prepare SOH data from log data", LogType.INFO)
        self.prepare_soh_data_from_log_data()

        if self.req_data_streams != ['*']:
            not_found_data_streams = [ds for ds in self.req_data_streams
                                      if ds not in self.found_data_streams]
            if not_found_data_streams != []:
                msg = (f"No data found for data streams: "
                       f"{', '.join(map(str, not_found_data_streams))}")
                self.processing_log.append((msg, LogType.WARNING))

        super().finalize_data()

    def read_folder(self, folder: str) -> None:
        """
        Loop all files in dir/list_of_rt130_paths to read for soh data,
            mass position data and
            index waveform data with filename and corresponding time range

        :param folder: absolute path to data set folder
        """
        count = 0

        total = 0
        if self.list_of_rt130_paths != []:
            folders = self.list_of_rt130_paths
            for folder in folders:
                total += sum([len(files) for _, _, files in os.walk(folder)])
        else:
            folders = [self.dir]
            total = sum([len(files) for _, _, files in os.walk(self.dir)])

        for folder in folders:
            for path, subdirs, files in os.walk(folder):
                for file_name in files:
                    if self.creator_thread.isInterruptionRequested():
                        raise ThreadStopped()
                    path2file = Path(path).joinpath(file_name)
                    if not validate_file(path2file, file_name):
                        continue
                    try:
                        if not self.read_reftek_130(path2file):
                            read_text(path2file, self.log_data['TEXT'])
                    except Exception:
                        fmt = traceback.format_exc()
                        self.track_info(f"Skip file {path2file} can't be read "
                                        f"due to error: {str(fmt)}",
                                        LogType.WARNING)

                    count += 1
                    if count % 50 == 0:
                        self.track_info(
                            f"Read {count} files/ {total}", LogType.INFO)
        self.total_datafile = count
        self.track_info(
            f'Skipped {total - count} invalid files.', LogType.INFO)

    def select_key(self) -> Tuple[str, str]:
        """
        :return selected_key:
            (device's serial number, experiment_number)
            the selected keys from available key of stream header.
        + If there is only one key, return it.
        + If there is more than one, show all keys, let user choose one to
        return.
        """
        keys = sorted(list(self.stream_header_by_key_chan.keys()))
        if len(keys) == 0:
            msg = 'No required data found for the data set.'
            raise ProcessingDataError(msg)

        selected_key = keys[0]
        if len(keys) > 1:
            msg = ("There are more than one keys in the given data.\n"
                   "Please select one to display")
            self.pause_signal.emit(msg, keys)
            self.pause()
            selected_key = keys[self.pause_response]

        self.track_info(f'Select Key {selected_key}', LogType.INFO)
        return selected_key

    def read_reftek_130(self, path2file: Path) -> bool:
        """
        From the given file:
            + Read SOH data from file with SH packets,
            + read event info, mass position and index waveform (data stream)
                file from file with EH or ET packets

        :param path2file: absolute path to file
        """
        try:
            rt130 = core.Reftek130.from_file(path2file)
        except Exception:
            fmt = traceback.format_exc()
            self.track_info(f"Skip file {path2file} can't be read "
                            f"due to error: {str(fmt)}", LogType.WARNING)
            return
        unique, counts = np.unique(rt130._data["packet_type"],
                                   return_counts=True)
        nbr_packet_type = dict(zip(unique, counts))

        if b"SH" in nbr_packet_type:
            self.read_sh_packet(path2file)
        if b"EH" in nbr_packet_type or b"ET" in nbr_packet_type:
            self.read_eh_or_et_packet(rt130)
        return True

    def read_sh_packet(self, path2file: Path) -> None:
        """
        Use soh_packet library to read file with SH packet for soh data
            to append tuple (time, log string) to log_data[self.cur_key][SOH]

        :param path2file: absolute path to file
        """
        with open(path2file, "rb") as fh:
            str = fh.read()
        data = soh_packet._initial_unpack_packets_soh(str)
        for ind, d in enumerate(data):
            cur_key = (d['unit_id'].decode(),
                       f"{d['experiment_number']}")
            self.populate_cur_key_for_all_data(cur_key)
            logs = soh_packet.SOHPacket.from_data(d).__str__()
            if 'SOH' not in self.log_data[cur_key]:
                self.log_data[cur_key]['SOH'] = []
            self.log_data[cur_key]['SOH'].append((d['time'], logs))

    def read_eh_or_et_packet(self, rt130: core.Reftek130) -> None:
        """
        Files that contents EH or ET packets are data stream (DS and
            mass position (DS 9) files.
        Name of file is name of packets' DS + 1.
        To be consistent with filename, check DS by the real value + 1
        The DS to be selected from GUI is also this value.
        There can be at most 1 - 9 DSs.

        DSs that aren't selected will be excluded first.
        Then data will be retrieved for log_data, mass_pos_data and
            waveform data.

        :param rt130: RT130 object of a data stream file.
        """
        if len(rt130._data) == 0:
            return
        data_stream = rt130._data['data_stream_number'][0] + 1
        if (self.req_data_streams != ['*'] and
                data_stream not in self.req_data_streams + [9]):
            return
        self.found_data_streams.append(data_stream)
        # suppose all data in a reftek file have the same key
        cur_key = (rt130._data[0]['unit_id'].decode(),
                   f"{rt130._data[0]['experiment_number']}")
        self.populate_cur_key_for_all_data(cur_key)
        if data_stream != 9:
            # don't get event info for mass position
            self.get_ehet_in_log_data(rt130, cur_key)
        self.get_mass_pos_data_and_waveform_data(rt130, data_stream, cur_key)

    def get_ehet_in_log_data(self, rt130: core.Reftek130,
                             cur_key: Tuple[str, str]) -> None:
        """
        Read event header info to add to log_data['EHET']

        :param rt130: RT130 object
        :param cur_key: current key of data set:
        """
        ind_ehet = [ind for ind, val in
                    enumerate(rt130._data["packet_type"])
                    if val in [b"EH"]]     # only need event header
        nbr_dt_samples = sum(
            [rt130._data[ind]["number_of_samples"]
             for ind in range(0, len(rt130._data))
             if ind not in ind_ehet])

        for index in ind_ehet:
            d = rt130._data[index]
            logs = core.EHPacket(d).eh_et_info(nbr_dt_samples)
            if 'EHET' not in self.log_data[cur_key]:
                self.log_data[cur_key]['EHET'] = []
            self.log_data[cur_key]['EHET'].append((d['time'], logs))

    def get_mass_pos_data_and_waveform_data(self, rt130: core.Reftek130,
                                            data_stream: int,
                                            cur_key: Tuple[str, str]) -> None:
        """
        Get mass_pos_data for the current key in DS 9.
        Get waveform_data for the current key in DS 1-8
        Read header to find the available trace indexes to decide to read
            deeply for real data.

        :param rt130: RT130 object
        :param data_stream: index of data stream (1-8)
        :param cur_key: current key of data set
        """
        if data_stream == 9:
            cur_data_dict = self.mass_pos_data[cur_key]
        elif self.rt130_waveform_data_req:
            cur_data_dict = self.waveform_data[cur_key]
        else:
            return

        avail_trace_indexes = check_reftek_header(
            rt130, cur_key, self.read_start, self.read_end,
            self.stream_header_by_key_chan,
            cur_data_dict, self.data_time[cur_key],
            self.include_mp123zne, self.include_mp456uvw)
        if avail_trace_indexes == []:
            return
        read_reftek_stream(
            rt130, self.tmp_dir, cur_key, avail_trace_indexes,
            cur_data_dict,
            self.include_mp123zne, self.include_mp456uvw)

    def prepare_soh_data_from_log_data(self) -> None:
        """
        Process SOH and event log_data to create SOHdata
        """
        for k in self.log_data:
            if self.creator_thread.isInterruptionRequested():
                raise ThreadStopped()
            if k == 'TEXT':
                continue
            if k not in self.data_time:
                self.data_time[k] = [constants.HIGHEST_INT, 0]
            self.soh_data[k] = {}
            logs = []
            for pkt_type in ['SOH', 'EHET']:
                if pkt_type == 'EHET':
                    # according to the above looping list, EHET will be
                    # at the bottom of log_str
                    logs += '\n\nEvents:'
                try:
                    # sort log data according to time (x[0])
                    self.log_data[k][pkt_type] = sorted(
                        self.log_data[k][pkt_type], key=lambda x: x[0])
                    for log in self.log_data[k][pkt_type]:
                        logs += log[1]
                        if pkt_type == 'SOH':
                            logs += '\n'
                except KeyError:
                    pass
            log_str = ''.join(logs)
            self.log_data[k] = {'SOH': [log_str]}
            log_obj = LogInfo(
                self, self.track_info, log_str, k,
                self.req_data_streams, self.req_soh_chans)
            self.data_time[k][0] = min(log_obj.min_epoch, self.data_time[k][0])
            self.data_time[k][1] = max(log_obj.max_epoch, self.data_time[k][1])
            for c_name in self.soh_data[k]:
                c = self.soh_data[k][c_name]['tracesInfo'][0]
                tr = {
                    'times': np.array(c['times']),
                    'data': np.array(c['data']),
                    'logIdx': np.array(c['logIdx']),
                    'startTmEpoch': self.data_time[k][0],
                    'endTmEpoch': self.data_time[k][1]
                }
                self.soh_data[k][c_name]['tracesInfo'] = [tr]
